import 'dart:ffi';
import 'dart:typed_data';
import 'package:ffi/ffi.dart' as ffi;
import 'package:ffi_helper/ffi_helper.dart';
import 'package:meta/meta.dart';

/// Stores the [ptr] to an array and its [length]
abstract class Array<T extends NativeType> {
  final Pointer<T> ptr;
  final int length;

  /// Takes ownership of [ptr] to an array which is [length] long.
  Array(this.ptr, this.length);

  /// Frees [ptr] to the array.
  void delete();

  @override
  String toString() {
    return '${runtimeType}($ptr, $length)';
  }
}

/// Basic implemenatation of [Array] using [ffi](https://pub.dev/packages/ffi)
/// for allocation and deallocation
class BasicArray<T extends NativeType> extends Array<T> {
  BasicArray(int length) : super(ffi.allocate(count: length), length);
  BasicArray.fromPtr(Pointer<T> ptr, int length) : super(ptr, length);

  @override
  void delete() {
    ffi.free(ptr);
  }
}

/// Non-owning [Array], meaning [delete] doesn't do anything.
class ArrayView<T extends NativeType> extends Array<T> {
  ArrayView.fromPtr(Pointer<T> ptr, int length) : super(ptr, length);
  ArrayView(Array<T> arr) : super(arr.ptr, arr.length);

  /// Doesn't do anything.
  @override
  void delete() {}
}

/// Takes an [Array] and exposes a [list] to its contents.
/// The exact type of list can be narrowed with extension methods like [Uint8ArrayList].
/// [list] becomes invalid after [array] got deleted.
class ArrayList<T extends NativeType> extends DelegatingArray<T> {
  @visibleForOverriding
  final List<int> list;

  /// [list] is a view into the contents of [arr.ptr]; see extension methods of [Pointer] at
  /// [dart:ffi](https://api.dart.dev/stable/2.10.5/dart-ffi/dart-ffi-library.html).
  /// To create an [ArrayList], prefer extension methods like [ToUint8Arraylist].
  ArrayList(Array<T> array, this.list) : super(array);
}

extension AsUint8ArrayList on Array<Uint8> {
  @factory
  ArrayList<Uint8> asArrayList() {
    return ArrayList(this, ptr.asTypedList(length));
  }
}

extension Uint8ArrayList on ArrayList<Uint8> {
  Uint8List get typedList => list;
}
